"""


Implementation of ML introductory algorithms visulisation as ancillary
Author - mr frozen (or just frozen)
Github - github.com/mrfrozen97
Mail - mrfrozenpeak@gmail.com



This file has class that performs Logistic regression algorithm to the input data and can also visualise it .
The packages used here are numpy(for efficient calculations) and matplotlib(for data visualisation)

This is a preliminary file that performs binary Logistic regression. This is one of the first versions. So a lot
will be improved in the future.


Classes:
Logistic Regression- It carries out regression model
plot model- It carries out data visualisation.







#Mathematics.........................................................................................................................

Maths behind this algorithm.

Equations
f(x) = y = w1x1 + w2x3 + w3x3 + ...... + b
(w1, w2, w3... are weights and b is bias)

We use Sigmoid function to normalise our values. It makes sure that our values lie between 0 and 1.
There is a term called as decision boundary. It is a threshold output(between 0 and 1) above which our point
belongs to a particular class and below which it belongs to other.

We used squared error cost function in linear regression, but in logistic regression while achieving gradient decent we
would rather choose a cost function which causes decent faster when we are far from minimum and slowly decends when we
move closer to the minimum.
So we use cost function called as cross entropy.


Cost function
When it belongs to class 1
=> -log( hø(x))
When it belongs to class 2
=> -log(1 - hø(x))


Or we can simply write it as,
Cost(hø(x,y)) = -y*log(hø(x)) - (1-y)*log(1 - hø(x))

We can get this formula by following derivation.
Let y' = p(y=1|x)   (It means that probability that y == 1 given the value of x)

so, 1-y' = p(y=0|x)

When y=0 , we have 1-y'  and when y=1 we have y'
So, we can generalise it as,

P(y|x) = y'^(y) * (1-y')^(1-y)
-Log(P(y|x)) = y*log(y') + (1-y)*log(1-y')
Which is denoted by -L(y',y)


z = w1x1 + w2x3 + w3x3 + ...... + b
y' = 1/(1 - e^-(z))


Now, using this we need to find what changes should we make in our weights and biases so that that next time we get
better results.
So, let changes to be made in weights be ∂W

∂L/∂W1 = ∂L/∂y' * ∂y'/∂z * ∂z/dW1

1st
∂L/∂y' = -y/y' + (1-y)/(1-y') = (-y(1-y') + y'(1-y)) / (y')(1-y')

2nd
∂y'/∂z = (y')(1-y')

3rd
∂z/∂W1 = x1

∂L/∂W1 = [(-y(1-y') + y'(1-y)) / (y')(1-y')]  * (y')(1-y') * x1
=> [yy' - y + y' - yy' ] * x1
=> (y' - y)*x1

Similarly, using similar concept
∂L/∂b = (y' - y)


Now since we calculated how much change in weights and biases..........
So, we can update them as follows...........

W1 = W1 - α*x1*(y' - y)
b = b - α*(y' - y)

Where, α is the learning rate or the speed at which we want to reach at the minimum.
It is important to not mess up with this rate, because very large learning rate can surpass/skip the minimum and we
won't get the best results and very small learning rate makes it too difficult to actually reach the minimum.

The above equation displays how we would update one weight, but real life data has n number of dimentions and so we need
to update each and every weight and there is only one bias independent of the number of xs because it is constant.



#End of math. Maybe, Never skip maths................................................................









Attributes:
............................
____________________________
x                  -> input x parameters (list of list or np array)
y                  -> input y for training data
showGrid           -> Boolean which indicates weather grid should be shown or not
sizeX              -> Horizontal size of graph in inches
sizeY              -> Vertical size of graph in inches
labelx_count       -> number of labels on x axis
labely_count       -> number of labels on y axis
title_color        -> color of title of plot
label_default      -> Boolean which tells if the label is default or changed by user
xlabel_color       -> color of x label
ylabel_color       -> color of y label
title              -> title text
x_label            -> List x label text for all parameters/dimentions
y_label            -> y label text
xlabel_size        -> x_label font size of text
ylabel_size        -> y_label font size of text
title_size         -> title font size of text
model              -> this is the model object i.e. logistic regression object in case we need any parent class variables
x_calculated       -> The calculated part of model's x parameters or the test data's x coordinates
y_calculated       -> Calculated classification using model or the test data
learning_rate      -> The rate at with we want to train our model (usually in order of 0.001)
n_iters            -> Number of iterations which we we want our model to pass through the training process
weights            -> List of weights used in our model
bias               -> single float/double containing our bias value
parameters         -> count of x parameters, dimentions
coordinates_size   -> size of training data
calculated_point_class0_label -> train data class 0 color
calculated_point_class1_label -> train data class 1 point color
calculated_point_class0_color -> test data class 0 point color
calculated_point_class1_color -> test data class 1 point color
calculated_point_alpha        -> alpha/transparency of the point


#####.....IMP.......####
display_graph -> Specify weather graph needs to be displayed before saving in plot_graph method
color_dict -> Dictionary that contains wide range of colors in the form of dictionary. The keys are the names of colors
              , values are the hex codes of that color. You can get this dict using get_color_dict function()
You can print this dict i.e. print(object.get_color_dict()) to know which colors are available.






You are most welcomed to improve this code:
You can pull request for this code at

github.com/mrfrozen97/                      (In spiderAlgorithms repo)
or
Email - mrfrozenpeak@gmail.com






"""








import numpy as np
import matplotlib.pyplot as plt
from frozenSpider import spiderAlgorithmResources as res
#from sklearn import datasets
#from sklearn.model_selection import  train_test_split
#from frozenSpider import Data_visualisation as dv





# Class which implements logistic regression...................................................................s

class LogisticRegression:



    #init method takes in some default input values. It is not compulsory to set this variables, everything is predefined
    #using default values......................

    def __init__(self, learning_rate=0.0001, n_iters = 1000 ):
        self.learning_rate = learning_rate
        self.n_iters = n_iters
        self.weights = None
        self.bias = None
        self.parameters = 0
        self.coordinates_size = 0
        self.x = []
        self.y = []
        self.y_actual = []
        self.x_calculated = []
        self.y_calculated = []


     #Function which calculates z.......
     #Math formula explained above code...................................


    def calculate_z(self, x_row):

        z = np.dot(x_row, self.weights) + self.bias
       # print(z)
        return z



    #Sigmoid normalisation function

    def sigmoid(self, z):

        return (1/(1+np.exp(-z)))





    #setter function to set decision boundary if any............................................

    def set_decision_boundary(self, x):
        self.decision_boundary = x





    #Following the comman naming convention best fit function calculated all weights and biases.......................

    def best_fit(self, x, y):
        self.x = x
        self.y = y
        x = np.array(x)

        self.coordinates_size, self.parameters = x.shape
        self.decision_boundary = 0.5
        self.weights = []
        self.bias = 3
        self.weights = np.zeros(self.parameters)


        for i in range(self.n_iters):
            alpha = self.sigmoid(self.calculate_z(x))
            dw = (1/self.coordinates_size) * np.dot(x.T, (alpha - y))
            db = (1/self.coordinates_size) * np.sum(alpha - y)

            self.weights -= self.learning_rate*dw
            self.bias -= self.learning_rate*db

            #print(dw, db)
        self.y_actual = alpha






    #Predict method used to test the test data/ unkown data.............................................


    def predict(self, x, decision_boundary=0.5):
        self.x_calculated = x
        self.decision_boundary = decision_boundary

        alpha = self.sigmoid(self.calculate_z(x))
        self.y_calculated = alpha
        calculated_class = [1 if i>=self.decision_boundary else 0 for i in alpha]
       # print(list(self.sigmoid(self.calculate_z(x))))
        return calculated_class







#This function plots n graphs of n dimentions of the logistic regression model


class plot_model(LogisticRegression):


    #All the default value are already set but they can all be reset.........................
    #Model object need to be passes so that we get other information about the model....................

    def __init__(self, model):

        self.showGrid = True
        self.sizeX = 10
        self.sizeY = 6
        self.labelx_count = 10
        self.labely_count = 10
        self.title_color = "#FF0000"
        self.xlabel_color = "#663399"
        self.ylabel_color = "#663399"
        self.label_default = True
        self.title = ""
        self.x_label = ["Dimention " + str(x) for x in range(1,100)]
        self.y_label = ""
        self.xlabel_size = 15
        self.ylabel_size = 15
        self.title_size = 18
        self.model = model
        self.color_dict = res.Resources.get_color_dict()
        self.calculated_point_size = 10
        self.calculated_point_class0_label = "Calculated class 0"
        self.calculated_point_class1_label = "Calculated class 1"
        self.calculated_point_class0_color = "red"
        self.calculated_point_class1_color = "cyan"
        self.calculated_point_alpha = 0.5







    #Function to get dictionary to access the different options avaliable for the colors.........................

    def get_color_dict(self):
        return self.color_dict








     #function to set label of each dimention of the graph................................................

    def set_dimention_labels(self, dimention_label):
        self.x_label = dimention_label






    #function to set vlues for title, x/y label and values of labels in case.........................

    def set_marker_properties(self,
                              label_default=True,
                              calculated_point_class1_label = "Calculated class 1",
                              calculated_point_class0_label="Calculated class 0",
                              calculated_point_alpha = 0.5,
                              calculated_point_class1_color='red'
                              ,calculated_point_class0_color="cyan",
                              title_size=15, xlabel_size=10,
                              ylabel_size=10, title="Output vs dimention",
                              y_label="y coordinates",
                              title_color="#FF0000",
                              xlabel_color="#663399",
                              ylabel_color="#663399"):



        self.title_size = title_size
        self.xlabel_size = xlabel_size
        self.ylabel_size = ylabel_size
        self.y_label = y_label
        self.title = title
        self.xlabel_color = xlabel_color
        self.ylabel_color = ylabel_color
        self.title_color = title_color
        self.label_default = label_default
        self.calculated_point_size = 10
        self.calculated_point_class0_label = calculated_point_class0_label
        self.calculated_point_class1_label = calculated_point_class1_label
        self.calculated_point_class0_color = calculated_point_class0_color
        self.calculated_point_class1_color = calculated_point_class1_color
        self.calculated_point_alpha = calculated_point_alpha










   #Sets the display size which will be equal to size of image if it is being svaed.........................

    def set_display_size(self, sizeX, sizeY):
        self.sizeX = sizeX
        self.sizeY = sizeY













    # main fuction thats plots the graph and saved it if path provided.........................
    #This function iterates through all the dimentions/parameters and plot each one seperate...............




    def plot_model(self,
                   display_graph = True,
                   display_calculated_points = True,
                   alpha=0.6, point_size=25,label_x=[], label_y=[],
                   class1_color='purple',
                   class0_color='orange',
                   unknown_points_label="Unknown points",
                   line_label='Best Fit line',
                   point_color='DeepSkyBlue',
                   save_fig_file='dont'):


        # These are the coordinates to plot the line which is inclusion of both known and calculated points


        multi_dimentions = np.array(self.model.x)
        multi_dimentions_calculated = np.array(self.model.x_calculated)
       # print(len(multi_dimentions[0]))

        for dimention in range(len(multi_dimentions[0])):



            x_bestFitLine_coords = multi_dimentions.T[dimention]
            y_bestFitLine_coords = self.model.y_actual

            x_bestFitLine_coords1 = []
            x_bestFitLine_coords2 = []
            y_bestFitLine_coords1 = []
            y_bestFitLine_coords2 = []
            for i in range(len(y_bestFitLine_coords)):
                if y_bestFitLine_coords[i]>=0.5:
                    x_bestFitLine_coords1.append(x_bestFitLine_coords[i])
                    y_bestFitLine_coords1.append(y_bestFitLine_coords[i])
                else:
                    x_bestFitLine_coords2.append(x_bestFitLine_coords[i])
                    y_bestFitLine_coords2.append(y_bestFitLine_coords[i])

            x_bestFitLine_coords_calculated = multi_dimentions_calculated.T[dimention]
            y_bestFitLine_coords_calculated = self.model.y_calculated

            x_bestFitLine_coords1_calculated = []
            x_bestFitLine_coords2_calculated = []
            y_bestFitLine_coords1_calculated = []
            y_bestFitLine_coords2_calculated = []
            for i in range(len(y_bestFitLine_coords_calculated)):
                if y_bestFitLine_coords_calculated[i] >= 0.5:
                    x_bestFitLine_coords1_calculated.append(x_bestFitLine_coords_calculated[i])
                    y_bestFitLine_coords1_calculated.append(y_bestFitLine_coords_calculated[i])
                else:
                    x_bestFitLine_coords2_calculated.append(x_bestFitLine_coords_calculated[i])
                    y_bestFitLine_coords2_calculated.append(y_bestFitLine_coords_calculated[i])




            plt.scatter(x_bestFitLine_coords1, y_bestFitLine_coords1, color=self.color_dict[class1_color],
                        label="class 1", alpha=alpha, zorder=3)                     #plot the best fit line
            plt.scatter(x_bestFitLine_coords2, y_bestFitLine_coords2, color=self.color_dict[class0_color],
                        label="class 0", alpha=alpha, zorder=3)  # plot the best fit line

            if display_calculated_points:
                plt.scatter(x_bestFitLine_coords1_calculated, y_bestFitLine_coords1_calculated, color=self.color_dict[self.calculated_point_class0_color],
                            label=self.calculated_point_class0_label, alpha=alpha, zorder=3)  # plot the best fit line
                plt.scatter(x_bestFitLine_coords2_calculated, y_bestFitLine_coords2_calculated, color=self.color_dict[self.calculated_point_class1_color],
                            label=self.calculated_point_class1_label, alpha=alpha, zorder=3)  # plot the best fit line



            plt.title(self.title,
                      fontdict=res.Resources.title_dict,
                      color=self.title_color)
            plt.xlabel(self.x_label[dimention],
                       fontdict=res.Resources.label_dict,
                       color=self.xlabel_color)
            plt.ylabel(self.y_label,
                       fontdict=res.Resources.label_dict,
                       color=self.ylabel_color)





            plt.legend(loc = 'upper right')

            if self.showGrid:
                plt.grid(color='#cfd8dc', zorder=0)

            figure = plt.gcf()
            figure.set_size_inches(self.sizeX, self.sizeY)


            if not(save_fig_file=='dont') :

                  plt.savefig("./"+save_fig_file +"/"+self.x_label[dimention])#, bbox_inches='tight')
            if display_graph:
               plt.show()
            plt.close()


















#######Sample code..............................................................................................................








#Example 1
"""
lor = LogisticRegression()
lor.best_fit([[1,2, 4],[2,3, 5],[4,5, 8]], [0, 0, 1])

cal = lor.predict([[1,0,4],[2,3, 5],[4,5, 8]])
print(cal)
"""









#Example 2
#Import sciketlearn before implementing it
"""
bc = datasets.load_breast_cancer()
x, y = bc.data, bc.target

x_train, x_test, y_train, y_test = train_test_split(x, y, test_size=0.2, random_state=1234)


lor = LogisticRegression(learning_rate=0.01, n_iters=10000)
lor.best_fit(x_train, y_train)
prediction = lor.predict(x_test)

accuracy = np.sum(prediction == y_test)/ len(prediction)

print(accuracy)

plot = plot_model(lor)
plot.set_marker_properties()
plot.plot_model(save_fig_file="./sample_graphs", display_graph=False)

"""





#Exampl2 extention. How to plot a T-sne 3D graph using above Logrithmic regression

"""
datav = dv.Logistic_regression_plot(lor)
datav.plot_3D_visuals(plot_test_data=True, save_fig_path="./sample_graphs")
"""



#Exampl2 extention. How to plot a T-sne 2D graph using above Logrithmic regression

"""
datav = dv.Logistic_regression_plot2D(lor)
datav.plot_2D_visuals(plot_test_data=True, save_fig_path="./sample_graphs")

"""











import matplotlib.pyplot as plt
import numpy as np



class SupportVectorMachine:

    def __init__(self):
        self.b = 0
        self.data = None
        self.w = []

    def fit(self, data):
        self.data = data

        opt_dict = {}
        transform = [[1, 1],
                      [-1, 1],
                      [1, -1],
                      [-1, -1]]
        all_data = []


        for yi in self.data:
            for featureset in self.data[yi]:
                for feature in featureset:
                    all_data.append(feature)

        self.max_feature_value = max(all_data)
        self.min_feature_value = min(all_data)
        all_data = None

        step_sizes = [self.max_feature_value * 0.1,
                      self.max_feature_value * 0.01,
                      self.max_feature_value * 0.001]

        b_range_multiple = 5
        b_multiple = 5


        latest_optimum = self.max_feature_value * 10



        for step in step_sizes:
            w = np.array([latest_optimum, latest_optimum])

            optimized = False
            while not optimized:
                for b in np.arange(-1*(self.max_feature_value*b_range_multiple),
                                   self.max_feature_value*b_range_multiple,
                                   step*b_multiple):
                    for transformation in transform:
                        w_t = w*transformation
                        found_option = True

                        for i in self.data:
                            for xi in self.data[i]:
                                yi=i
                                if not yi*(np.dot(w_t, xi)+b) >= 1:
                                    found_option =False
                        if found_option:
                            opt_dict[np.linalg.norm(w_t)] = [w_t, b]
                if w[0] < 0:
                    optimized = True
                    print("Optimised a step...")

                else:
                    w = w - step
            norms = sorted([n for n in opt_dict])
            opt_choice = opt_dict[norms[0]]

            self.w = opt_choice[0]
            self.b = opt_choice[1]
            latest_optimum = opt_choice[0][0] + step*2


    def predict(self, x):
        classification = np.sign(np.dot(np.array(x, self.w)) + self.b)
        plt.scatter(x[0], x[1])

        return classification

    def visualize(self):

        plt.style.use('dark_background')
        for i in data_dict:

            for x in data_dict[i]:
                plt.scatter(x[0], x[1])

        def hyperplane(x, w, b, v):
            return (-w[0]*x-b+v)/ w[1]

        datarange = (self.min_feature_value*0.9, self.max_feature_value*1.1)
        hyp_x_min = datarange[0]
        hyp_x_max = datarange[1]

        psv1 = hyperplane(hyp_x_min, self.w, self.b, 1)
        psv2 = hyperplane(hyp_x_max, self.w, self.b, 1)
        plt.plot([hyp_x_min, hyp_x_max], [psv1, psv2])

        nsv1 = hyperplane(hyp_x_min, self.w, self.b, -1)
        nsv2 = hyperplane(hyp_x_max, self.w, self.b, -1)
        plt.plot([hyp_x_min, hyp_x_max], [nsv1, nsv2])

        db1 = hyperplane(hyp_x_min, self.w, self.b, 0)
        db2 = hyperplane(hyp_x_max, self.w, self.b, 0)
        plt.plot([hyp_x_min, hyp_x_max], [db1, db2])

        plt.show()










data_dict = {-1:np.array([[1,7],[2,8],[3,8]]), 1:np.array([[5,1], [6,-1], [7,3]])}


svm = SupportVectorMachine()
svm.fit(data_dict)
svm.visualize()



from frozenSpider import Data_visualisation as dv
from sklearn import datasets as ds


dataset = ds.load_breast_cancer()
x, y = dataset.data, dataset.target

dv.plot_parallel_plot(x, y,                                                        # Compulsory parameters

                      title="Breat cancer Dataset",                                # Optional parameters
                      x_label="Parameters",
                      y_label="Values",
                      alpha=0.4,
                      class_colors=['Red', 'Cyan'],                          # Class colors can be any color name There are 145 options
                      plot_labels=['0 class', '1 class'])







"""


This file implements K-means cluster algorithm. This is a simple but powerful algorithm.

This file has class that  can also visualise it .

This is a preliminary file that performs K-means cluster. This is one of the first versions. So a lot
will be improved in the future.


Classes:
K_means_cluster- It carries out model
plot model- It carries out data visualisation.




Attributes:
............................
____________________________
x                  -> input x parameters (list of list or np array)
y                  -> input y for training data
showGrid           -> Boolean which indicates weather grid should be shown or not
sizeX              -> Horizontal size of graph in inches
sizeY              -> Vertical size of graph in inches
labelx_count       -> number of labels on x axis
labely_count       -> number of labels on y axis
title_color        -> color of title of plot
label_default      -> Boolean which tells if the label is default or changed by user
xlabel_color       -> color of x label
ylabel_color       -> color of y label
title              -> title text
x_label            -> List x label text for all parameters/dimentions
y_label            -> y label text
xlabel_size        -> x_label font size of text
ylabel_size        -> y_label font size of text
title_size         -> title font size of text
model              -> this is the model object i.e. logistic regression object in case we need any parent class variables
x_calculated       -> The calculated part of model's x parameters or the test data's x coordinates
y_calculated       -> Calculated classification using model or the test data
learning_rate      -> The rate at with we want to train our model (usually in order of 0.001)
n_iters            -> Number of iterations which we we want our model to pass through the training process
weights            -> List of weights used in our model
bias               -> single float/double containing our bias value
parameters         -> count of x parameters, dimentions
coordinates_size   -> size of training data
calculated_point_class0_label -> train data class 0 color
calculated_point_class1_label -> train data class 1 point color
calculated_point_class0_color -> test data class 0 point color
calculated_point_class1_color -> test data class 1 point color
calculated_point_alpha        -> alpha/transparency of the point


#####.....IMP.......####
display_graph -> Specify weather graph needs to be displayed before saving in plot_graph method
color_dict -> Dictionary that contains wide range of colors in the form of dictionary. The keys are the names of colors
              , values are the hex codes of that color. You can get this dict using get_color_dict function()
You can print this dict i.e. print(object.get_color_dict()) to know which colors are available.






You are most welcomed to improve this code:
You can pull request for this code at

github.com/mrfrozen97/                      (In spiderAlgorithms repo)
or
Email - mrfrozenpeak@gmail.com




"""

import matplotlib.pyplot as plt
import numpy as np
import random
#from sklearn import datasets
#from sklearn.model_selection import train_test_split
from matplotlib.animation import FuncAnimation
#from sklearn.manifold import TSNE



class K_means_cluster():

    def __init__(self):

        self.x =[]
        self.y = []
        self.centeroids = []
        self.k = 2
        self.n_iters = 300
        self.clusters = {}
        self.unknown = []
        self.cluster_population = []
        self.graph_centeroids = []
        self.graph_centeroids2D = []
        self.plot_x = []
        self.plot_y = []



    def fit(self, x, y, k, n_iters=300):
        self.x = x
        self.y = y
        self.k = k
        self.centeroids = []
        self.n_iters = n_iters

        """
        i = 0
        while i<k:
            rand_centeroid = random.choice(self.x)
            if rand_centeroid in self.centeroids:
                pass
            else:
                i+=1
                self.centeroids.append(rand_centeroid)
                
        """
        for _ in range(k):
            self.centeroids.append(self.x[random.randint(0,len(self.x)-1)])
            self.cluster_population.append({})

        #print(self.centeroids)



        for i in range(self.n_iters):

            for j in range(k):
                self.clusters[j] = []
                self.cluster_population[j] = {}


            for index,coord in enumerate(self.x):
                dist_list = []
                for centeroid in self.centeroids:
                    dist_list.append(np.sum(np.square(np.subtract(coord, centeroid))))

                temp_index=dist_list.index(min(dist_list))
                self.clusters[temp_index].append(coord)
                if self.y[index] in self.cluster_population[temp_index]:
                    self.cluster_population[temp_index][self.y[index]] +=1
                else:
                    self.cluster_population[temp_index][self.y[index]] = 1



            #print(self.clusters)

            prev_centereoids = self.centeroids
            new_centeroids = []

            for cluster in self.clusters:

                temp_centeroid = np.average(self.clusters[cluster], axis=0)
                #print(temp_centeroid)
                new_centeroids.append(list(temp_centeroid))

                #print(temp_centeroid[0], temp_centeroid[1])

            self.centeroids = new_centeroids
            self.graph_centeroids.append(prev_centereoids)

            if np.sum(np.abs(np.subtract(new_centeroids, prev_centereoids))) < 0.0001:

                print("Optimised: " + str(i))
                break

        #print(self.cluster_population)
        #print(list(self.graph_centeroids))


    def get_centeroid_transition(self):

        return self.graph_centeroids

    def get_clusters(self):

        return self.cluster_population



    def predict(self, unknown):

        self.unknown = unknown
        calculated_unknown = []
        for coord in self.unknown:
            dist_list = []
            for centeroid in self.centeroids:
                dist_list.append(np.sum(np.square(np.subtract(coord, centeroid))))

            temp_index = dist_list.index(min(dist_list))
            temp_index_class = max(self.cluster_population[temp_index], key=self.cluster_population[temp_index].get)
            calculated_unknown.append(temp_index_class)
        return calculated_unknown







    def animate(self, i):
        self.plot_x.append(self.graph_centeroids[i][0])
        self.plot_y.append(self.graph_centeroids[i][1])


    def graph_animations(self):
        print(self.graph_centeroids)



        self.plot_x.append(self.graph_centeroids[0][0])
        self.plot_y.append(self.graph_centeroids[0][1])

        plt.scatter(self.plot_x, self.plot_y)
        ani = FuncAnimation(plt.gcf(), self.animate, interval=2000, frames=2)
        plt.show()















#Example
"""

coords = [
         [2, 3],
         [7, 11],
         [-1, 0],
         [5, 6],
         [7, 10],
         [2, 2.5],
         [9, 10.4],
         [4, 5],
         [5.5, 3.8],
         [4.7, 3.2],
         [7.5, 6.2]
          ]
y = [0, 1, 0, 1, 0, 1, 1, 0, 1, 0, 1]



#datasetb = datasets.load_breast_cancer()
#x, y = datasetb.data, datasetb.target

#x_train, x_test, y_train, y_test = train_test_split(x, y, test_size=0.2, random_state=1234)


model = K_means_cluster()
model.fit(coords, y, 2, n_iters=100)

#res = model.predict(x_test)



count = 0

#for i in range(len(res)):
 #   if res[i]==y_test[i]:
 #       count+=1


print(model.get_clusters())


#print("Accuracy = " + str(count/len(res)*100))

"""


from setuptools import setup, find_packages


classifiers = [
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Education',
    'Operating System :: Microsoft :: Windows :: Windows 10',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3'
]

setup(
    name='frozenSpider',
    version='0.0.5',
    description='Library to implement and visualise ML Algorithms',
    long_description=open('ReadMe.txt').read() + '\n\n' + open('CHANGELOG.txt').read(),
    url='https://github.com/mrfrozen97',
    author='mr_frozen97',
    author_email='mrfrozenpeak@gmail.com',
    license="MIT",
    classifiers=classifiers,
    keywords='ml algorithms',
    packages=find_packages(),
    install_requires=['numpy']
)


import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
from frozenSpider.spiderAlgorithmResources import Resources as res
import random

class Linear_Regression_Animation():

    def __init__(self, x, y):

        self.plot_x = []
        self.plot_y = []
        self.curr_b = 0
        self.curr_a = 0
        self.x = x
        self.y = y
        self.x_sum = 0
        self.xy_sum = 0
        self.y_sum = 0
        self.xx_sum = 0
        self.max_x = max(self.x)
        self.min_x = min(self.x)
        self.points_color = 'blue'
        self.line_color = 'red'
        self.color_dict = res.get_color_dict()
        self.alpha = 0.6
        self.x_ticks_count = 15





    def set_plot_properties(self, alpha=0.6, x_ticks_count=15, line_color='red', points_color='blue'):

        self.alpha = alpha
        self.line_color = line_color
        self.points_color = points_color
        self.x_ticks_count = x_ticks_count



    def calculate_coords(self, i):

        self.plot_x.append(self.x[i])
        self.plot_y.append(self.y[i])

        self.x_sum += self.x[i]
        self.y_sum += self.y[i]
        self.xx_sum += self.x[i]**2
        self.xy_sum += self.x[i]*self.y[i]

        numerator = self.xy_sum/(i+1) - (self.y_sum/(i+1) * self.x_sum/(i+1))
        denominator = self.xx_sum/(i+1) - (self.x_sum/(i+1))**2

        if denominator!= 0:
            self.curr_b = numerator/denominator
            self.curr_a = self.y_sum/(i+1) - self.curr_b*self.x_sum/(i+1)
        else:
            self.curr_b = numerator
            self.curr_a = self.y_sum / (i + 1) - self.curr_b * self.x_sum / (i + 1)
        plt.clf()
        plt.scatter(self.plot_x, self.plot_y, color=self.color_dict[self.points_color], alpha=self.alpha)
        plt.plot([self.min_x, self.max_x] ,[self.min_x*self.curr_b + self.curr_a, self.max_x*self.curr_b + self.curr_a],
                 color=self.color_dict[self.line_color])
        #plt.yticks(self.y[::10])
        plt.xticks(self.x[::int(len(self.x)/(len(self.x)/self.x_ticks_count))])
        plt.title("a = " + str(self.curr_a) + "  b = " + str(self.curr_b))
        #plt.style.use('dark_background')



        #print(self.curr_b, end=" ")
        #print(self.curr_a)


    def get_color_dict(self):

        return res.get_color_dict()



    def animate(self, interval=300, frames='defalut'):

        plt.yticks(self.y)
        plt.xticks(self.x)
        #plt.style.use('dark_background')
        plt.scatter(self.plot_x, self.plot_y)
        if frames =='defalut':
            frames = len(self.x)-1
        ani = FuncAnimation(plt.gcf(), self.calculate_coords, interval=interval, frames=frames)
        plt.show()












x = [1, 2, 3, 4, 5, 6, 7, 8, 9]
y = [3, 5, 7.5, 9.4, 10.6, 12.5, 15, 17.6, 20]

x = [i for i in range(200)]
y = [2*i+random.random()*random.randint(-5,6)*i**(0.5) for i in x]


animate = Linear_Regression_Animation(x, y)
animate.animate()












import numpy as np
import nnfs
from nnfs.datasets import spiral_data
nnfs.init()
# Dense layer
class Layer_Dense :
    # Layer initialization
    def __init__ ( self , n_inputs , n_neurons ):
        # Initialize weights and biases
        self.weights = 0.01 * np.random.randn(n_inputs, n_neurons)
        self.biases = np.zeros(( 1 , n_neurons))
        # Forward pass
    def forward ( self , inputs ):
        # Remember input values
        self.inputs = inputs
        # Calculate output values from inputs, weights and biases
        self.output = np.dot(inputs, self.weights) + self.biases
    # Backward pass
    def backward ( self , dvalues ):
        # Gradients on parameters
        self.dweights = np.dot(self.inputs.T, dvalues)
        self.dbiases = np.sum(dvalues, axis = 0 , keepdims = True )
        # Gradient on values
        self.dinputs = np.dot(dvalues, self.weights.T)

# ReLU activation
class Activation_ReLU :
    # Forward pass
    def forward ( self , inputs ):

        # Remember input values
        self.inputs = inputs
        # Calculate output values from inputs
        self.output = np.maximum( 0 , inputs)
        # Backward pass
    def backward ( self , dvalues ):
        # Since we need to modify original variable,
        # let's make a copy of values first
        self.dinputs = dvalues.copy()
        # Zero gradient where input values were negative
        self.dinputs[self.inputs <= 0 ] = 0






class Layer_Dropout :
    # Init
    def __init__ ( self , rate ):
        # Store rate, we invert it as for example for dropout
        # of 0.1 we need success rate of 0.9
        self.rate = 1 - rate
    # Forward pass
    def forward ( self , inputs ):
        # Save input values
        self.inputs = inputs
        # Generate and save scaled mask
        self.binary_mask = np.random.binomial( 1 , self.rate,
        size = inputs.shape) / self.rate
        # Apply mask to output values
        self.output = inputs * self.binary_mask
    # Backward pass
    def backward ( self , dvalues ):
        # Gradient on values
        self.dinputs = dvalues * self.binary_mask








# Softmax activation
class Activation_Softmax :
    # Forward pass
    def forward ( self , inputs ):
        # Remember input values
        self.inputs = inputs
        # Get unnormalized probabilities
        exp_values = np.exp(inputs - np.max(inputs, axis = 1 ,keepdims = True ))
        # Normalize them for each sample
        probabilities = exp_values / np.sum(exp_values, axis = 1 ,keepdims = True )
        self.output = probabilities
    # Backward pass
    def backward ( self , dvalues ):
        # Create uninitialized array
        self.dinputs = np.empty_like(dvalues)
        # Enumerate outputs and gradients
        for index, (single_output, single_dvalues) in enumerate ( zip (self.output, dvalues)):
        # Flatten output array
            single_output = single_output.reshape( - 1 , 1 )
            # Calculate Jacobian matrix of the output and
            jacobian_matrix = np.diagflat(single_output) - np.dot(single_output, single_output.T)

        # Calculate sample-wise gradient
        # and add it to the array of sample gradients
        self.dinputs[index] = np.dot(jacobian_matrix,
        single_dvalues)



# Common loss class
class Loss :
    # Calculates the data and regularization losses
    # given model output and ground truth values
    def calculate ( self , output , y ):
        # Calculate sample losses
        sample_losses = self.forward(output, y)
        # Calculate mean loss
        data_loss = np.mean(sample_losses)
        # Return loss
        return data_loss
        # Cross-entropy loss

class Loss_CategoricalCrossentropy ( Loss ):
    # Forward pass
    def forward ( self , y_pred , y_true ):
        # Number of samples in a batch
        samples = len (y_pred)
        # Clip data to prevent division by 0
        # Clip both sides to not drag mean towards any value
        y_pred_clipped = np.clip(y_pred, 1e-7 , 1 - 1e-7 )
        # Probabilities for target values -
        # only if categorical labels
        if len (y_true.shape) == 1 :
            correct_confidences = y_pred_clipped[
            range (samples),
            y_true
            ]

        # Mask values - only for one-hot encoded labels
        elif len (y_true.shape) == 2 :
            correct_confidences = np.sum(
            y_pred_clipped * y_true,
            axis = 1
            )
        # Losses
        negative_log_likelihoods = - np.log(correct_confidences)
        return negative_log_likelihoods
    # Backward pass
    def backward ( self , dvalues , y_true ):
        # Number of samples
        samples = len (dvalues)
        # Number of labels in every sample
        # We'll use the first sample to count them
        labels = len (dvalues[ 0 ])
        # If labels are sparse, turn them into one-hot vector
        if len (y_true.shape) == 1 :
            y_true = np.eye(labels)[y_true]
        # Calculate gradient
        self.dinputs = - y_true / dvalues
        # Normalize gradient
        self.dinputs = self.dinputs / samples



# Softmax classifier - combined Softmax activation
# and cross-entropy loss for faster backward step
class Activation_Softmax_Loss_CategoricalCrossentropy ():
    # Creates activation and loss function objects
    def __init__ ( self ):
        self.activation = Activation_Softmax()
        self.loss = Loss_CategoricalCrossentropy()
    # Forward pass
    def forward ( self , inputs , y_true ):
        # Output layer's activation function
        self.activation.forward(inputs)
        # Set the output
        self.output = self.activation.output
        # Calculate and return loss value
        return self.loss.calculate(self.output, y_true)

    # Backward pass
    def backward ( self , dvalues , y_true ):
        # Number of samples
        samples = len (dvalues)
        # If labels are one-hot encoded,
        # turn them into discrete values
        if len (y_true.shape) == 2 :
            y_true = np.argmax(y_true, axis = 1 )
        # Copy so we can safely modify
        self.dinputs = dvalues.copy()
        # Calculate gradient
        self.dinputs[ range (samples), y_true] -= 1
        # Normalize gradient
        self.dinputs = self.dinputs / samples






# SGD optimizer
class Optimizer_SGD :
    # Initialize optimizer - set settings,
    # learning rate of 1. is default for this optimizer
    def __init__ ( self , learning_rate = 1. , decay = 0. , momentum = 0. ):
        self.learning_rate = learning_rate
        self.current_learning_rate = learning_rate
        self.decay = decay
        self.iterations = 0
        self.momentum = momentum

    # Call once before any parameter updates
    def pre_update_params ( self ):
        if self.decay:
            self.current_learning_rate = self.learning_rate * \
            ( 1. / ( 1. + self.decay * self.iterations))
    # Update parameters
    def update_params ( self , layer ):
    # If we use momentum
        if self.momentum:
            # If layer does not contain momentum arrays, create them
            # filled with zeros
            if not hasattr (layer, 'weight_momentums' ):
                layer.weight_momentums = np.zeros_like(layer.weights)
                # If there is no momentum array for weights
                # The array doesn't exist for biases yet either.
                layer.bias_momentums = np.zeros_like(layer.biases)
            # Build weight updates with momentum - take previous
            # updates multiplied by retain factor and update with
            # current gradients
            weight_updates = \
            self.momentum * layer.weight_momentums - \
            self.current_learning_rate * layer.dweights
            layer.weight_momentums = weight_updates
            # Build bias updates
            bias_updates = \
            self.momentum * layer.bias_momentums - \
            self.current_learning_rate * layer.dbiases
            layer.bias_momentums = bias_updates
            # Vanilla SGD updates (as before momentum update)
        else :
            weight_updates = - self.current_learning_rate * \
            layer.dweights
            bias_updates = - self.current_learning_rate * \
            layer.dbiases
        # Update weights and biases using either
        # vanilla or momentum updates
        layer.weights += weight_updates
        layer.biases += bias_updates

    # Call once after any parameter updates
    def post_update_params ( self ):
        self.iterations += 1



# Adagrad optimizer
class Optimizer_Adagrad :
    # Initialize optimizer - set settings
    def __init__ ( self , learning_rate = 1. , decay = 0. , epsilon = 1e-7 ):
        self.learning_rate = learning_rate
        self.current_learning_rate = learning_rate
        self.decay = decay
        self.iterations = 0
        self.epsilon = epsilon
    # Call once before any parameter updates
    def pre_update_params ( self ):
        if self.decay:
            self.current_learning_rate = self.learning_rate * \
            ( 1. / ( 1. + self.decay * self.iterations))
        # Update parameters

    def update_params ( self , layer ):
        # If layer does not contain cache arrays,
        # create them filled with zeros
        if not hasattr (layer, 'weight_cache' ):
            layer.weight_cache = np.zeros_like(layer.weights)
            layer.bias_cache = np.zeros_like(layer.biases)
            # Update cache with squared current gradients
            layer.weight_cache += layer.dweights ** 2
            layer.bias_cache += layer.dbiases ** 2
            # Vanilla SGD parameter update + normalization
            # with square rooted cache
            layer.weights += - self.current_learning_rate * \
            layer.dweights / \
            (np.sqrt(layer.weight_cache) + self.epsilon)
            layer.biases += - self.current_learning_rate * \
            layer.dbiases / \
            (np.sqrt(layer.bias_cache) + self.epsilon)
    # Call once after any parameter updates
    def post_update_params ( self ):
        self.iterations += 1


# RMSprop optimizer
class Optimizer_RMSprop :
    # Initialize optimizer - set settings
    def __init__ ( self , learning_rate = 0.001 , decay = 0. , epsilon = 1e-7 ,
    rho = 0.9 ):
        self.learning_rate = learning_rate
        self.current_learning_rate = learning_rate
        self.decay = decay
        self.iterations = 0
        self.epsilon = epsilon
        self.rho = rho
    # Call once before any parameter updates
    def pre_update_params ( self ):
        if self.decay:
            self.current_learning_rate = self.learning_rate * \
            ( 1. / ( 1. + self.decay * self.iterations))
    # Update parameters
    def update_params ( self , layer ):
        # If layer does not contain cache arrays,
        # create them filled with zeros
            if not hasattr (layer, 'weight_cache' ):
                layer.weight_cache = np.zeros_like(layer.weights)
                layer.bias_cache = np.zeros_like(layer.biases)
            # Update cache with squared current gradients
            layer.weight_cache = self.rho * layer.weight_cache + \
            ( 1 - self.rho) * layer.dweights ** 2
            layer.bias_cache = self.rho * layer.bias_cache + \
            ( 1 - self.rho) * layer.dbiases ** 2
            # Vanilla SGD parameter update + normalization
            # with square rooted cache
            layer.weights += - self.current_learning_rate * \
            layer.dweights / \
            (np.sqrt(layer.weight_cache) + self.epsilon)
            layer.biases += - self.current_learning_rate * \
            layer.dbiases / \
            (np.sqrt(layer.bias_cache) + self.epsilon)
    # Call once after any parameter updates
    def post_update_params ( self ):
        self.iterations += 1




# Adam optimizer
class Optimizer_Adam :
    # Initialize optimizer - set settings
    def __init__ ( self , learning_rate = 0.001 , decay = 0. , epsilon = 1e-7 ,
    beta_1 = 0.9 , beta_2 = 0.999 ):
        self.learning_rate = learning_rate
        self.current_learning_rate = learning_rate
        self.decay = decay
        self.iterations = 0
        self.epsilon = epsilon
        self.beta_1 = beta_1
        self.beta_2 = beta_2
    # Call once before any parameter updates
    def pre_update_params ( self ):
        if self.decay:
            self.current_learning_rate = self.learning_rate * ( 1. / ( 1. + self.decay * self.iterations))
    # Update parameters
    def update_params ( self , layer ):
        # If layer does not contain cache arrays,
        # create them filled with zeros
        if not hasattr (layer, 'weight_cache' ):
            layer.weight_momentums = np.zeros_like(layer.weights)
            layer.weight_cache = np.zeros_like(layer.weights)
            layer.bias_momentums = np.zeros_like(layer.biases)
            layer.bias_cache = np.zeros_like(layer.biases)
        # Update momentum with current gradients
        layer.weight_momentums = self.beta_1 * layer.weight_momentums + ( 1 - self.beta_1) * layer.dweights
        layer.bias_momentums = self.beta_1 * layer.bias_momentums + ( 1 - self.beta_1) * layer.dbiases
        # Get corrected momentum
        # self.iteration is 0 at first pass
        # and we need to start with 1 here
        weight_momentums_corrected = layer.weight_momentums /( 1 - self.beta_1 ** (self.iterations + 1 ))
        bias_momentums_corrected = layer.bias_momentums /( 1 - self.beta_1 ** (self.iterations + 1 ))
        # Update cache with squared current gradients
        layer.weight_cache = self.beta_2 * layer.weight_cache + ( 1 - self.beta_2) * layer.dweights ** 2

        layer.bias_cache = self.beta_2 * layer.bias_cache + ( 1 - self.beta_2) * layer.dbiases ** 2
        # Get corrected cache
        weight_cache_corrected = layer.weight_cache / ( 1 - self.beta_2 ** (self.iterations + 1 ))
        bias_cache_corrected = layer.bias_cache / ( 1 - self.beta_2 ** (self.iterations + 1 ))
        # Vanilla SGD parameter update + normalization
        # with square rooted cache
        layer.weights += - self.current_learning_rate * weight_momentums_corrected /(np.sqrt(weight_cache_corrected) +
        self.epsilon)
        layer.biases += - self.current_learning_rate * bias_momentums_corrected / (np.sqrt(bias_cache_corrected) +self.epsilon)

    # Call once after any parameter updates
    def post_update_params ( self ):
        self.iterations += 1
















# Create dataset
X, y = spiral_data( samples = 100 , classes = 3 )
# Create Dense layer with 2 input features and 3 output values
dense1 = Layer_Dense( 2 , 64 )
# Create ReLU activation (to be used with Dense layer):
activation1 = Activation_ReLU()
# Create second Dense layer with 3 input features (as we take output
# of previous layer here) and 3 output values (output values)
dense2 = Layer_Dense( 64 , 3 )
# Create Softmax classifier's combined loss and activation
loss_activation = Activation_Softmax_Loss_CategoricalCrossentropy()
# Perform a forward pass of our training data through this layer




# Create optimizer
optimizer = Optimizer_Adam( learning_rate = 0.05 , decay = 5e-7 )




for epoch in range ( 10001 ):
    # Perform a forward pass of our training data through this layer
    dense1.forward(X)
    # Perform a forward pass through activation function
    # takes the output of first dense layer here
    activation1.forward(dense1.output)
    # Perform a forward pass through second Dense layer
    # takes outputs of activation function of first layer as inputs
    dense2.forward(activation1.output)
    # Perform a forward pass through the activation/loss function
    # takes the output of second dense layer here and returns loss
    loss = loss_activation.forward(dense2.output, y)
    # Calculate accuracy from output of activation2 and targets
    # calculate values along first axis
    predictions = np.argmax(loss_activation.output, axis = 1 )
    if len (y.shape) == 2 :
        y = np.argmax(y, axis = 1 )
    accuracy = np.mean(predictions == y)
    if not epoch % 100 :
        print ('epoch:' +str(epoch) + ' acc:' + str(accuracy)+ ' loss:'+ str(loss) + 'lr:'+ str(optimizer.current_learning_rate))
    # Backward pass
    loss_activation.backward(loss_activation.output, y)
    dense2.backward(loss_activation.dinputs)
    activation1.backward(dense2.dinputs)
    dense1.backward(activation1.dinputs)
    # Update weights and biases
    optimizer.pre_update_params()
    optimizer.update_params(dense1)
    optimizer.update_params(dense2)
    optimizer.post_update_params()







"""

dense1.forward(X)
# Perform a forward pass through activation function
# takes the output of first dense layer here
activation1.forward(dense1.output)
# Perform a forward pass through second Dense layer
# takes outputs of activation function of first layer as inputs
dense2.forward(activation1.output)
# Perform a forward pass through the activation/loss function
# takes the output of second dense layer here and returns loss
loss = loss_activation.forward(dense2.output, y)


# Let's see output of the first few samples:
print (loss_activation.output[: 5 ])
# Print loss value
print ( 'loss:' , loss)
# Calculate accuracy from output of activation2 and targets
# calculate values along first axis
predictions = np.argmax(loss_activation.output, axis = 1 )
if len (y.shape) == 2:
    y = np.argmax(y, axis = 1 )
accuracy = np.mean(predictions == y)



# Print accuracy
print ( 'acc:', accuracy)
# Backward pass
loss_activation.backward(loss_activation.output, y)
dense2.backward(loss_activation.dinputs)
activation1.backward(dense2.dinputs)
dense1.backward(activation1.dinputs)
# Print gradients
print (dense1.dweights)
print (dense1.dbiases)
print (dense2.dweights)
print (dense2.dbiases)


"""



